# -*- coding: utf-8 -*-
#***************************************************************************
#    Copyright (C) 2010 by Eike Welk                                       *
#    eike.welk@gmx.net                                                     *
#                                                                          *
#    License: GPL                                                          *
#                                                                          *
#    This program is free software; you can redistribute it and/or modify  *
#    it under the terms of the GNU General Public License as published by  *
#    the Free Software Foundation; either version 2 of the License, or     *
#    (at your option) any later version.                                   *
#                                                                          *
#    This program is distributed in the hope that it will be useful,       *
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#    GNU General Public License for more details.                          *
#                                                                          *
#    You should have received a copy of the GNU General Public License     *
#    along with this program; if not, write to the                         *
#    Free Software Foundation, Inc.,                                       *
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
#***************************************************************************

"""
Test code for the "interpreter.py" module
    High level tests involving program text in Siml.
"""

from __future__ import division
from __future__ import absolute_import              

from py.test import skip as skip_test # pylint: disable-msg=F0401,E0611,W0611
from py.test import fail as fail_test # pylint: disable-msg=F0401,E0611,W0611

from freeode.util import assert_raises #pylint:disable-msg=W0611 



def test_data_statement_simple_1(): #IGNORE:C01111
    msg = 'Test data statement: create attributes'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, IFloat, IString
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
data a: Float const
data b: String const
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    
    assert isinstance(mod.a, IFloat)
    assert isinstance(mod.b, IString)
  
  

def test_data_statement_roles_1(): #IGNORE:C01111
    msg = 'Test data statement: create attributes with different roles'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, isrole
    from freeode.ast import RoleConstant, RoleParameter, RoleVariable
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
data a: Float const
data b: Float param
data c: Float variable
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print mod
    
    assert isrole(mod.a, RoleConstant)
    assert isrole(mod.b, RoleParameter)
    assert isrole(mod.c, RoleVariable)
    
  

def test_data_statement_roles_2(): #IGNORE:C01111
    '''
    In user defined classes the role should be propagated recursively to the 
    child attributes.
    However, the child attributes' roles are not simply replaced, but only 
    changed if they are more variable than the new role. 
    
    These are the main roles, the variable character increases from left to 
    right:
    RoleConstant, RoleParameter, RoleVariable, RoleUnkown
    const,        param,         variable,     role_unknown
    '''
    msg = 'Test data statement: roles should be propagated to child attributes.'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, isrole
    from freeode.ast import RoleConstant, RoleParameter, RoleVariable
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
class A:
    data c: Float const
    data p: Float param
    data v: Float variable
    
data ac: A const
data ap: A param
data av: A variable
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
#    
#    print
#    print 'attribute ac: ---------------------------------'
#    print aa_make_tree(mod.ac)
    
    #all attributes should be const
    assert isrole(mod.ac.c, RoleConstant)
    assert isrole(mod.ac.p, RoleConstant)
    assert isrole(mod.ac.v, RoleConstant)
  
    #Roles: c: const; p,v: param
    assert isrole(mod.ap.c, RoleConstant)
    assert isrole(mod.ap.p, RoleParameter)
    assert isrole(mod.ap.v, RoleParameter)
  
    #Roles: c: const; p: param; v: variable
    assert isrole(mod.av.c, RoleConstant)
    assert isrole(mod.av.p, RoleParameter)
    assert isrole(mod.av.v, RoleVariable)
  
  

def test_builtin_function_call_1(): #IGNORE:C01111
    msg = 'Test interpreter object: call built in function sqrt'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 
    import math
    
    prog_text = \
'''
data a: Float const
a = sqrt(2)
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
    
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    
    assert mod.a.value == math.sqrt(2)
  
  

def test_builtin_function_call_2(): #IGNORE:C01111
    msg = 'Test interpreter object: call built in function printc'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
printc('test')
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(intp.modules['test'])
  
  

def test_function_definition_1(): #IGNORE:C01111
    msg = '''Test all legal styles of function definitions.'''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 

    prog_text = \
'''
func foo1(a, b):
    return a + b
    
func foo2(a=1, b=2):
    return a + b
    
func foo3(a:Float, b:Float):
    return a + b

func foo4(a:Float=1, b:Float=2):
    return a + b
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
    
    _mod = intp.modules['test']
  
    #TODO: write assertions
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
  
  

def test_function_call_1(): #IGNORE:C01111
    msg = '''Test all legal styles of function calls.'''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 

    prog_text = \
'''
func foo(a:Float=1, b:Float=2):
    return a + b

data a,b,c,d,e,f,g,h: Float const

a = foo()           # a == 3       - line 7
b = foo(5)          # b = 7
c = foo(5, 6)       # c = 11
d = foo(a=10)       # d = 12       - line 10
e = foo(b=20)       # e = 21
f = foo(5, b=20)    # f = 25
g = foo(a=10, b=20) # g = 30
h = foo(b=20, a=10) # h = 30       - line 14
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
    
    #get the interpreted module
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    
    #test the results
    assert mod.a.value == 3
    assert mod.b.value == 7
    assert mod.c.value == 11
    assert mod.d.value == 12
    assert mod.e.value == 21
    assert mod.f.value == 25
    assert mod.g.value == 30
    assert mod.h.value == 30
 
  

def test_function_definition_and_call_1(): #IGNORE:C01111
    msg = 'Interpreter: slightly complex function definition and function call.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 

    prog_text = \
'''
printc('start')

func foo(b):
    printc('in foo. b = ', b)
    return b*b
    printc('after return')

data a: Float const
a = 2*2 + foo(3*4) + foo(2)
printc('a = ', a)

printc('end')
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    
    assert mod.a.value == 2*2 + (3*4)**2 + 2**2
  
  

def test_printc_function_1(): #IGNORE:C01111
    msg = 'Test the printc function. - actual printing: Float, String, expression.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    
    prog_text = \
'''
#print known constants
printc(23)
printc('hello ',2, ' the world!')

#print unknown value
data foo: Float
data bar: String
printc(foo)
printc(bar)

#print unevaluated expression
data a,b: Float variable
printc(a+b)
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')



def test_print_function_2(): #IGNORE:C01111
    msg = 'Test printc function. - compile time printing of user defined class.'
    #skip_test(msg)
    print msg
    #TODO: This does not work as intended. Either refine infrastructure or delete test case.
    #      See bug #596679 
    #        https://bugs.launchpad.net/freeode/+bug/596679

    from freeode.interpreter import Interpreter
    
    prog_text = \
'''
#print user defined class
class C:
    data a: Float const
    data b: String const
    
    func __siml_str__(this):
        return a.__str__() + ' and ' + b.__str__()
    
data c: C
c.a = 5
c.b = 'hello'
printc(c)
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
  

def test_print_function_3(): #IGNORE:C01111
    msg = '''
    Test the print function. - code generation for: Float, String
    
    TODO: Broken! This is bug #597234
            https://bugs.launchpad.net/freeode/+bug/597234
    '''
    skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import DotName, aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
class A:
    data a,b,foo: Float
    data bar: String
    
    func dynamic(this):
        #print known constants
        print(23)
        print('hello ',2, ' the world!')
        
        #print unknown value
        print(foo)
        print(bar)
        
        #print unevaluated expression
        print(a+b)
        
compile A
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    _mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(_mod) 

    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #print sim
    #get the dynamic function with the generated code
    dynamic = sim.get_attribute(DotName('dynamic'))
    assert len(dynamic.statements) == 5



def test_print_function_4(): #IGNORE:C01111
    msg = 'Test the print function. - code generation for: user defined class.'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import DotName, aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
#print user defined class
class C:
    data a: Float 
    data b: String
    
    func __siml_str__(this):
        return a.__siml_str__() + ' and ' + b.__siml_str__()
        #return ' and ' + b.__siml_str__()


class A:
    data c: C
    
    func dynamic(this):
        print(c)
        
compile A
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
#    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod) 

    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #print sim
    
    #get the dynamic function with the generated code
    dynamic = sim.get_attribute(DotName('dynamic'))
    assert len(dynamic.statements) == 1
  
  

def test_graph_function_1(): #IGNORE:C01111
    msg = 'Test the print function. - code generation for: user defined class.'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import DotName, aa_make_tree #pylint:disable-msg=W0612 
    
    prog_text = \
'''
class A:
    data c: Float
    
    func final(this):
        graph(c)
        
compile A
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
#    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod) 

    #get flattened object
    sim = intp.get_compiled_objects()[0]
#    print aa_make_tree(sim) 
    
    #get the final function with the generated code
    final = sim.get_attribute(DotName('final'))
    assert len(final.statements) == 1
  
  

def test_interpreter_class_definition_1(): #IGNORE:C01111
    msg = 'Test interpreter: class definition and instantiation.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree #pylint:disable-msg=W0612 

    prog_text = \
'''
printc('start')

data pi: Float const
pi = 3.1415

class A:
    printc('in A definition')
    data a1: Float const
    data a2: Float const

class B:
    data b1: Float const
    b1 = pi
    data b2: Float const

data a: A const
data b: B const

a.a1 = 1
a.a2 = 2 * b.b1
printc('a.a1: ', a.a1, ', a.a2: ', a.a2)

printc('end')
'''

    #create the interpreter
    intp = Interpreter()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
  
    assert mod.pi.value == 3.1415
    assert mod.a.a1.value == 1
    assert mod.a.a2.value == 2 * 3.1415
    assert mod.b.b1.value == 3.1415
  
  
  
#def test_interpreter_class_definition_2(): #IGNORE:C01111
#    '''
#    Test user defined classes - correctness of parent attribute.
#    
#    Data attributes are copied when a class is instantiated. (All attributes
#    are constructed when the class is defined.) The parent pointers, which 
#    point back to the object that contains each object, must be updated 
#    by the copy algorithm.  Otherwise they point to the old parents before the 
#    copy.
#    
#    Interpreter Object has a __deepcopy__ function that takes care of this.
#    '''
#    msg = 'Test user defined classes - correctness of parent attribute.'
#    skip_test(msg)
#    print msg
#    
#    from freeode.interpreter import Interpreter
#
#    prog_text = \
#'''
#class A:
#    data z: Float const
#
#class B:
#    data a: A const
#
#
#data b:B const
#'''
#
#    #create the interpreter
#    intp = Interpreter()
#    intp.interpret_module_string(prog_text, None, 'test')
#  
#    print
#    print intp.modules['test']
#    
#    #get the instance objects defined in this program
#    mod = intp.modules['test']
#    b = mod.get_attribute('b')
#    a = b.get_attribute('a')
#    z = a.get_attribute('z')
#    
#    #check the correctness of the parent attributes
#    assert b.parent() is mod
#    assert a.parent() is b
#    assert z.parent() is a



def test_interpreter_method_call(): #IGNORE:C01111
    msg = 'Test interpreter: method call.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''
printc('start')

class A:
    data a1: Float const
    data a2: Float const
    
    func compute(this, x):
        printc('in compute x=', x)
        return x + 2
        
data a: A const
a.a1 = a.compute(3)

printc('a.a1 = ', a.a1)

printc('end')
'''

    #create the interpreter
    intp = Interpreter()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
  
    assert mod.a.a1.value == 5



def test_method_call_this_namespace_1(): #IGNORE:C01111
    msg = 'Test interpreter: method call, this namespace'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''
printc('start')

class A:
    data a1: Float const
    data a2: Float const
    
    func compute(this, x):
        printc('in compute x=', x)
        a1 = x
        a2 = a1 + 2
        
data a: A const
a.compute(3)

printc('a.a1 = ', a.a1)
printc('a.a2 = ', a.a2)

printc('end')
'''

    #create the interpreter
    intp = Interpreter()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
  
    assert mod.a.a1.value == 3
    assert mod.a.a2.value == 5

      
      
def test_method_call_this_namespace_2(): #IGNORE:C01111
    msg = 'Test interpreter: method call, this namespace'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''
class A:
    data a1: Float const
    data a2: Float const
    
    func compute(this, x):
        printc('in compute x=', x)
        a1 = x
        a2 = a1 + 2
        
class B:
    data a: A
    
    func compute(this, x):
        a.compute(x)
    
data b: B const
b.compute(3)

printc('b.a.a1 = ', b.a.a1)
printc('b.a.a2 = ', b.a.a2)
'''

    #create the interpreter
    intp = Interpreter()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
  
    assert mod.b.a.a1.value == 3
    assert mod.b.a.a2.value == 5

      
      
def test_Interpreter_assign_const_1(): #IGNORE:C01111
    msg = 'Test interpreter object: assignment. needs working data statement and number'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    
    prog_text = \
'''
data a: Float const
a = 2
'''

    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
    mod = intp.modules['test']
#    print
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    
    assert mod.a.value == 2
  
  

# -------- Test interpreter object - emit code ----------------------------------------
def test_Interpreter_assign_emit_code_1(): #IGNORE:C01111
    msg = 'Test Interpreter.assign: emit code without the usual infrastructure.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter, IFloat, istype
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.ast import NodeAssignment, NodeFuncCall

    prog_text = \
'''
data a: Float variable
data b: Float variable
b = 2*a #emit this statement
'''

    #create the interpreter
    intp = Interpreter()
    #enable collection of statements for compilation
    intp.start_collect_code()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
    #get the results of the collection process
    stmts, _locals = intp.stop_collect_code()
  
    mod = intp.modules['test']
#    print
#    print '--------------- main module ----------------------------------'
#    print aa_make_tree(mod)
#    print
#    print '--------------- collected statements ----------------------------------'
#    print aa_make_tree(stmts)
        
    #one statement: b = 2*a    
    assert len(stmts) == 1
    assert isinstance(stmts[0], NodeAssignment)                #  b = 2 * a
    assert isinstance(stmts[0].target, IFloat)                 #  b
    assert stmts[0].target is mod.b                            #  b
    assert stmts[0].target.value is None                       #  b
    assert isinstance(stmts[0].expression, NodeFuncCall)       #  2*a
    assert istype(stmts[0].expression, IFloat)                 #  2*a
    assert isinstance(stmts[0].expression.arguments[0], IFloat)#  2
    assert stmts[0].expression.arguments[0].value == 2         #  2
    assert isinstance(stmts[0].expression.arguments[1], IFloat)#  a
    assert stmts[0].expression.arguments[1] is mod.a           #  a
    assert stmts[0].expression.arguments[1].value is None      #  a
      


def test_Interpreter_assign_emit_code_2(): #IGNORE:C01111
    msg = '''Test Interpreter.assign and partial evaluation. 
             Emit code without the usual infrastructure.'''
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter, IFloat, istype
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.ast import NodeAssignment, NodeFuncCall


    prog_text = \
'''
data a: Float const
data b: Float variable
data c: Float variable
a = 2*2 #constant no statement emitted
b = 2+a #emit b = 8; compute 2*a at compile time
c = 2-b #emit everything
#printc('a = ', a)
#printc('b = ', b)
#printc('c = ', c)
'''

    #create the interpreter
    intp = Interpreter()
    #enable collection of statements for compilation
    intp.start_collect_code()
    #interpret the program
    intp.interpret_module_string(prog_text, None, 'test')
    #get the results of the collection process
    stmts, _locals = intp.stop_collect_code() 
 
    mod = intp.modules['test']
#    print
#    print '--------------- main module ----------------------------------'
#    print aa_make_tree(mod)
#    print
#    print '--------------- collected statements ----------------------------------'
#    print aa_make_tree(stmts)
        
    assert len(stmts) == 2
    # b = 4
    assert isinstance(stmts[0], NodeAssignment)
    assert isinstance(stmts[0].expression, IFloat) # 8
    assert stmts[0].expression.value == 6
    assert stmts[0].target is mod.b
    # c = 2*b
    assert isinstance(stmts[1], NodeAssignment)
    assert isinstance(stmts[1].expression, NodeFuncCall)        # 2 * b
    assert istype(stmts[1].expression, IFloat)
    assert isinstance(stmts[1].expression.arguments[0], IFloat) # 2
    assert stmts[1].expression.arguments[0].value == 2
    assert isinstance(stmts[1].expression.arguments[1], IFloat) # b
    assert stmts[1].expression.arguments[1].value is None   
    assert stmts[1].target is mod.c   
      


def test_exec_NodeCompileStmt_1(): #IGNORE:C01111
    msg = 'Simple compilation test. \n'\
          'Test Interpreter.visit_NodeCompileStmt'
    #skip_test(msg)
    print msg
    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.interpreter import (Interpreter, IFloat, SimlFunction)
    from freeode.util import DotName

    prog_text = \
'''
class A:
    data b: Float
     
    func dynamic(this):
        b = 2

compile A
'''

    #create the interpreter
    intp = Interpreter()
    #run program
    intp.interpret_module_string(prog_text, None, 'test')
  
    _mod = intp.modules['test']
    comp_obj = intp.get_compiled_objects()[0]
    #print aa_make_tree(_mod)
    #print aa_make_tree(comp_obj)

    #there must be one compiled object present
    assert len(intp.get_compiled_objects()) == 1
    
    
    #the attributes b and dynamic must exist
    assert isinstance(comp_obj.get_attribute(DotName('b')), IFloat)
    assert isinstance(comp_obj.get_attribute(DotName('dynamic')), SimlFunction)
    assert len(comp_obj.get_attribute(DotName('dynamic')).statements) == 1
    
    
    
    
def test_interpreter_user_defined_operators_1(): #IGNORE:C01111
    '''
    User defined operators must be converted into multiple statements,
    that contain only operations on fundamental types. Intermediate results
    are kept in function local variables, that are stored by the compile 
    statement.
    
    The used Siml class simulates a geometric vector class.
    '''
    msg = 'Test user defined operators - code generation.'
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter, IFloat, SimlFunction
    from freeode.util import DotName
    from freeode.util import func #, aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''                                     #1
class Vec1D:
    data x: Float role_unknown

    func __add__(this, other):          #5
        data res: Vec1D
        res.x = x + other.x
        return res
        
    func __siml_assign__(this, other):  #10
        x = other.x


class A:
    data a, b, c: Vec1D                 #15
    
    func dynamic(this):
        #--- invoke the operators ----
        c = a + b
                                        #20

compile A
'''

    #create the interpreter and interpret the mini-program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    #print intp.modules['test']
    #print intp.get_compiled_objects()[0] 
    
    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #get the attributes that we have defined
    a_x = sim.get_attribute(DotName('a.x'))
    b_x = sim.get_attribute(DotName('b.x'))
    c_x = sim.get_attribute(DotName('c.x'))
    dynamic = sim.get_attribute(DotName('dynamic'))
    #test some facts about the attributes
    assert isinstance(a_x, IFloat)
    assert isinstance(b_x, IFloat)
    assert isinstance(c_x, IFloat)
    assert isinstance(dynamic, SimlFunction)
    
    assert len(dynamic.statements) == 2
    stmt0 = dynamic.statements[0]
    stmt1 = dynamic.statements[1]
    #first statement (res.x = x + other.x) assigns to function local variable
    assert stmt0.target is not a_x 
    assert stmt0.target is not b_x 
    assert stmt0.target is not c_x
    assert stmt0.expression.function is func(IFloat.__add__)
    #second statement (c=a+b) assigns to c.x
    assert stmt1.target is c_x
    #second statement assigns temporary result of previous computation to attribute c.x
    assert stmt0.target is stmt1.expression
    #assert False



def test_interpreter_user_defined_operators_2(): #IGNORE:C01111
    '''
    User defined operators must also work with constant data. 
    Same class as in previous test, but all variables are constant.
    
    The used Siml class simulates a geometric vector class.
    '''
    msg = 'Test user defined operators - compatibility with const keyword.'
    #skip_test(msg)
    print msg
    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.interpreter import Interpreter, isrole 
    from freeode.ast import RoleConstant

    prog_text = \
'''
class Vec1D:
    data x: Float role_unknown

    func __add__(this, other):
        data res: Vec1D
        res.x = x + other.x
        return res
        
    func __siml_assign__(this, other):
        x = other.x


data a,b,c: Vec1D const
a.x = 2
b.x = 3


#--- invoke the operators ----
c = a + b
'''

    #create the interpreter and interpret the mini-program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    
    mod = intp.modules['test']
    #print
    #print 'Interpreted module: -----------------------------------------------------'
    #print mod
    
    #get the attributes that we have defined
    a_x = mod.a.x
    assert isrole(a_x, RoleConstant)
    assert a_x.value == 2
    b_x = mod.b.x
    assert isrole(b_x, RoleConstant)
    assert b_x.value == 3
    c_x = mod.c.x
    assert isrole(c_x, RoleConstant)
    assert c_x.value == 5
    


def test_interpreter_expression_statement_1(): #IGNORE:C01111
    msg = '''Test expression statement - code generation. 
    Unused expressions also generate code. The interpreter does not know about 
    pure functions and functions with side effects. 
    TODO: Calls to pure functions where the result is unused, could be optimized away.'''
    #skip_test(msg)
    print msg
    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.interpreter import Interpreter, IFloat, SimlFunction
    from freeode.ast import NodeExpressionStmt
    from freeode.util import DotName, func

    prog_text = \
'''
class A:
    data a,b: Float
    
    func dynamic(this):
        1+2
        a+b

compile A
'''

    #create the interpreter and interpret the mini-program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    #print intp.modules['test']
    #print intp.get_compiled_objects()[0] 
    
    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #get the attributes that we have defined
    a = sim.get_attribute(DotName('a'))
    b = sim.get_attribute(DotName('b'))
    dynamic = sim.get_attribute(DotName('dynamic'))
    #test some facts about the attributes
    assert isinstance(a, IFloat)
    assert isinstance(b, IFloat)
    assert isinstance(dynamic, SimlFunction)
    #only one statement is collected (a+b)
    assert len(dynamic.statements) == 1
    stmt0 = dynamic.statements[0]
    assert isinstance(stmt0, NodeExpressionStmt)
    assert stmt0.expression.function is func(IFloat.__add__)
    #assert False



def test_user_defined_class_roles_1(): #IGNORE:C01111
    '''
    The role keywords (const, param, variable, ...) should work with user 
    defined classes too.
    '''
    msg = 'Test user defined classes with different roles.'
    #skip_test(msg)
    print msg
    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.interpreter import Interpreter, isrole
    from freeode.ast import RoleConstant, RoleAlgebraicVariable
    from freeode.util import DotName
    
    prog_text = \
'''
class A:
    data a: Float

#use the class as a constant
data ac: A const
ac.a = 2

class B:
    #use the class as a variable
    data av: A 
    data v: Float
    
    func dynamic(this):
        av.a = v
        
compile B
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
  
#    print
    mod = intp.modules['test']
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)

    ac_a = mod.ac.a
    assert isrole(ac_a, RoleConstant) 

    #get flattened object
    sim = intp.get_compiled_objects()[0] 
#    print 'Flattened object: ---------------------------------'
#    print aa_make_tree(sim)
    av_a = sim.get_attribute(DotName('av.a'))
    assert isrole(av_a, RoleAlgebraicVariable)
  
  

def test_function_return_value_roles_1(): #IGNORE:C01111
    '''
    User defined functions can be called from constant and from variable 
    environments. Test the roles of their return values.
    This test only involves fundamental types.
    '''
    msg = 'Test roles of return values of user defined functions.'
    #skip_test(msg)
    print msg
    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 
    from freeode.interpreter import Interpreter, isrole
    from freeode.ast import RoleConstant, RoleAlgebraicVariable
    from freeode.util import DotName
    
    prog_text = \
'''
func plus2(x):
    data r: Float
    r = x + 2
    return r

data ac,bc: Float const
ac = 3
bc = plus2(ac)


class B:
    data av,bv: Float
    
    func dynamic(this):
        bv = plus2(av)
        
compile B
'''
    #create the interpreter
    intp = Interpreter()
    #run mini program
    intp.interpret_module_string(prog_text, None, 'test')
#    print
    mod = intp.modules['test']
#    print 'module after interpreter run: ---------------------------------'
#    print aa_make_tree(mod)
    ac = mod.ac
    bc = mod.bc
    assert isrole(ac, RoleConstant)
    assert isrole(bc, RoleConstant) 
    assert ac.value == 3
    assert bc.value == 5

#    #get flattened object
    sim = intp.get_compiled_objects()[0] 
#    print 'Flattened object: ---------------------------------'
#    print aa_make_tree(sim)
    av = sim.get_attribute(DotName('av'))
    bv = sim.get_attribute(DotName('bv'))
    assert isrole(av, RoleAlgebraicVariable) 
    assert isrole(bv, RoleAlgebraicVariable)
  
  

def test_interpreter_dollar_operator_1(): #IGNORE:C01111
    msg = 'Test "$" operator. Basic capabililities.'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, IFloat, SimlFunction, isrole
    from freeode.ast import RoleStateVariable, RoleTimeDifferential
    from freeode.util import DotName
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''
class A:
    data a1: Float 

    func dynamic(this):
        $a1 = a1

compile A
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    #print aa_make_tree(intp.modules['test'])
    #print aa_make_tree(intp.get_compiled_objects()[0])
    
    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #get the attributes that we have defined
    a1 = sim.get_attribute(DotName('a1'))
    a1_dt = sim.get_attribute(DotName('a1$time'))   #implicitly defined by $ operator
    dynamic = sim.get_attribute(DotName('dynamic'))
    
    #test some facts about the attributes
    assert isinstance(a1, IFloat)             #a1 is state variable, because it 
    assert isrole(a1, RoleStateVariable)      #has derivative
    assert isinstance(a1_dt, IFloat)     
    assert isrole(a1_dt, RoleTimeDifferential)# $a1 is a time differential
    assert isinstance(dynamic, SimlFunction)
    
    #test if assignment really is 'a1$time' = 'a1'
    assign = dynamic.statements[0]
    assert assign.target is a1_dt
    assert assign.expression is a1



def test_interpreter_dollar_operator_2(): #IGNORE:C01111
    msg = '''
    Test "$" operator. 
    Bug: $ operator did not work with attributes of user defined classes.
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, SimlFunction, IFloat, isrole
    from freeode.ast import RoleStateVariable, RoleTimeDifferential
    from freeode.util import DotName    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data z: Float
    
    func dynamic(this):
        $z = z

class B:
    data a: A

    func dynamic(this):
        a.dynamic()

compile B
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    #print aa_make_tree(intp.modules['test'])
    #print aa_make_tree(intp.get_compiled_objects()[0])

    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    #get the attributes that we have defined
    az = sim.get_attribute(DotName('a.z'))
    az_dt = sim.get_attribute(DotName('a.z$time'))   #implicitly defined by $ operator
    dynamic = sim.get_attribute(DotName('dynamic'))
    #test some facts about the attributes
    assert isinstance(az, IFloat)        #a1 is state variable, because it 
    assert isrole(az, RoleStateVariable) #has derivative
    assert isinstance(az_dt, IFloat)     
    assert isrole(az_dt, RoleTimeDifferential) # $a1 is time differential
    assert isinstance(dynamic, SimlFunction)
    
    #test if assignment really is 'a1$time' = 'a1'
    assign = dynamic.statements[0]
    assert assign.target is az_dt
    assert assign.expression is az
    #assert False

  
  
def test_compile_statement__small_simulation_program(): #IGNORE:C01111
    msg = 'Test compile statement: see if small simulation program can be compiled.'
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
data g: Float const
g = 9.81


class BarrelWithHole:
    data V, h: Float
    data A_bott, A_o, mu, q: Float param

    func dynamic(this):                            #10
        h = V/A_bott
        $V = q - mu*A_o*sqrt(2*g*h)
#        print('h: ', h)

    func initialize(this, q_in):
        V = 0;
        A_bott = 1; A_o = 0.02; mu = 0.55;
        q = q_in #0.05
 
                                                   #20
class RunTest:
    data system: BarrelWithHole

    func dynamic(this):
        system.dynamic()

    func initialize(this):
        system.initialize(0.55)
        solution_parameters(100, 1)
                                                   #30
    func final(this):
        graph(system.V, system.h)
        print('Simulation finished successfully.')
        

compile RunTest
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    #print aa_make_tree(intp.modules['test'])
    #print aa_make_tree(intp.get_compiled_objects()[0])
    
  
  
def test_compile_statement_1(): #IGNORE:C01111
    msg = '''
    Test the compile statement 
    - Flattening and storage of functions' local variables'''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, IFloat, SimlFunction
    from freeode.util import DotName    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data a1: Float 
    
    func foo(this, x):
        return x
        
    func dynamic(this):
        data b,c: Float
        b = foo(a1)
        c = foo(a1 + b)
        $a1 = b

compile A
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    #print intp.modules['test']
    #print intp.get_compiled_objects()[0] 
    
    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    
    #get the attributes that we have defined
    a1 = sim.get_attribute(DotName('a1'))
    a1_dt = sim.get_attribute(DotName('a1$time'))
    dynamic = sim.get_attribute(DotName('dynamic'))
    #test some facts about the attributes
    assert isinstance(a1, IFloat)
    assert isinstance(a1_dt, IFloat)
    assert isinstance(dynamic, SimlFunction)

    #check number of attributes, most are automatically generated
    #global variable:   time
    #methods:           initialize, dynamic, final, 
    #instance variables:   a1, $a1, 
    #local variables:      A.dynamic.b, A.dynamic.c, 
    #intermediate result:  A.foo.x, (2nd call)
    assert len(sim.attributes) == 9
    


def test_compile_statement_2(): #IGNORE:C01111
    msg = \
    '''
    Test the compile statement:
    - Additional initialization methods must be recognized, and put into 
      the flat object.
    - These methods have names of the form: init_xxx(this, ...)
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, IFloat, SimlFunction
    from freeode.ast import NodeAssignment
    from freeode.util import DotName    
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data a: Float 
    data b: Float param
    
    #This is the additional initialization function.
    func init_b(this, in_b):
        b = in_b #set parameter
        a = 0    #set initial value
        
    func initialize(this):
        b = 0.1 #set parameter
        a = 0   #set initial value
        
    func dynamic(this):
        $a = b

    func final(this):
        pass
        
compile A
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    #print aa_make_tree(intp.modules['test'])
    #print aa_make_tree(intp.get_compiled_objects()[0])
    
    #get flattened object
    sim = intp.get_compiled_objects()[0] 
    
    #get the attributes that we have defined
    a = sim.get_attribute(DotName('a'))
    a_dt = sim.get_attribute(DotName('a$time'))
    initialize = sim.get_attribute(DotName('initialize'))
    init_b = sim.get_attribute(DotName('init_b'))
    dynamic = sim.get_attribute(DotName('dynamic'))
    final = sim.get_attribute(DotName('final'))
    #test some facts about the attributes
    assert isinstance(a, IFloat)
    assert isinstance(a_dt, IFloat)
    assert isinstance(initialize, SimlFunction)
    assert isinstance(dynamic, SimlFunction)
    assert isinstance(final, SimlFunction)
    #test the additional initialization method 'init_b' a little closer
    assert isinstance(init_b, SimlFunction)
    #'init_b' must contain 2 assignments
    init_b_stmts = init_b.statements
    assert len(init_b_stmts) == 2
    assert isinstance(init_b_stmts[0], NodeAssignment)
    assert isinstance(init_b_stmts[1], NodeAssignment)
    #'init_b' must have 2 arguments: 'this', 'in_b'
    init_b_args = init_b.siml_signature.arguments
    assert len(init_b_args) == 2
    assert init_b_args[0].name == 'this'
    assert init_b_args[1].name == 'in_b'
    #check number of attributes, most are automatically generated
    #global variable:     time
    #methods:             initialize, init_b, dynamic, final, 
    #instance variables:  a, $a, b
    #local variables:     -
    assert len(sim.attributes) == 8
    #check number of external inputs. Should be one: 'in_b'
    assert len(sim.external_inputs) == 1
    


def test_pass_statement_1(): #IGNORE:C01111
    msg = '''
    Test the pass statement. Try the normal use case.
    - The pass statement does nothing it is necessary to create empty 
    functions and classes. 
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import (Interpreter, istype, SimlFunction, IFloat)
    from freeode.util import DotName    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
#empty class body
class Dummy:
    pass

data d: Dummy const


#empty function body
func f_dummy():
    pass
    
f_dummy()
    
    
#call class with pass statement in compiled code
class A:
    data x: Float 
    data d: Dummy
    
    func foo(this):
        pass
        
    func dynamic(this):
        foo()
        $x = 1

compile A
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    
    #test the module a bit
    mod = intp.modules['test']
    #print mod

    assert istype(mod.d, mod.Dummy)
    assert isinstance(mod.f_dummy, SimlFunction)
    assert isinstance(mod.A, type)

    #Test the compiled object
    flat_A = intp.get_compiled_objects()[0] 
    #print flat_A
    x = flat_A.get_attribute(DotName('x'))
    x_time = flat_A.get_attribute(DotName('x$time'))
    assert isinstance(x, IFloat)
    assert isinstance(x_time, IFloat)    
    #check number of attributes, most are automatically generated
    #global variable:     time
    #methods:             initialize, dynamic, final, 
    #instance variables:  x,  x$time
    assert len(flat_A.attributes) == 6
    


def test_pass_statement_2(): #IGNORE:C01111
    msg = '''
    Test the pass statement. Try stupid but legal cases.
    - The pass statement should just do nothing. 
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, istype, IFloat, SimlFunction    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
#class with definition after pass statement
class A:
    pass
    data x: Float role_unknown

data a:A
a.x = 2

#function with statement after pass statement
func add2(x):
    pass
    return x + 2
    
data four: Float const
four = add2(2)
'''

    #create the interpreter
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
  
    print
    
    #test the module a bit
    mod = intp.modules['test']
    #print aa_make_tree(mod)
    
    A = mod.A
    a = mod.a
    add2 = mod.add2
    four = mod.four

    assert isinstance(A, type)
    assert istype(a, A)
    assert isinstance(a.x, IFloat)
    assert a.x.value == 2
    assert isinstance(add2, SimlFunction)
    assert isinstance(four, IFloat)
    assert four.value == 4



def test_ifc_statement_1(): #IGNORE:C01111
    msg = '''
    Test the "ifc" statement. Just constant code, no code creation.
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 

    prog_text = \
'''
data a,b: Float const
a = 2 

ifc a == 1:
    b = 1
elif a == 2:
    b = 2
else:
    b = 3
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')

    
    #test the module a bit
    mod = intp.modules['test']
    #print aa_make_tree(mod)
    
    assert mod.a.value == 2
    assert mod.b.value == 2
    
    
    
def test_ifc_statement_2(): #IGNORE:C01111
    msg = '''
    Test the "ifc" statement. Code generation is enabled. However all conditions 
    evaluate to false. No statements are generated inside the dynamic function.
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter
    from freeode.util import DotName    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data a,b: Float
    
    func dynamic(this):       
        ifc 0 == 1:
            b = 1
        elif 0 == 2:
            b = 2
        elif 0 == 3:
            b = 3

compile A
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')

    #the module
    #mod = intp.modules['test']
    #print mod
    
    sim = intp.get_compiled_objects()[0]
    #print sim
    #look at generated function 
    dynamic = sim.get_attribute(DotName('dynamic'))
    assert len(dynamic.statements) == 0 #function has no statements




def test_if_statement_1(): #IGNORE:C01111
    msg = '''
    Test the if statement. Code is generated and condition involves variables.
    All branches of the statement must be visited.
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import (Interpreter, isrole, IFloat, IBool)
    from freeode.ast import NodeIfStmt, NodeAssignment, RoleConstant
    from freeode.util import DotName    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data a,b: Float
    
    func dynamic(this):       
        if a == 1:
            b = 1
        elif a == 2:
            b = 2
        else:
            b = 3

compile A
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')

    #the module
    #mod = intp.modules['test']
    #print mod
    
    sim = intp.get_compiled_objects()[0]
    #print sim
    #look at variables
    a = sim.get_attribute(DotName('a'))
    b = sim.get_attribute(DotName('b'))
    #a,b are RoleVariable, so the asignments are not done at runtime
    assert a.is_known == False
    assert b.is_known == False
    #look at generated function and if statement
    dynamic = sim.get_attribute(DotName('dynamic'))
    assert len(dynamic.statements) == 1 #function has one statement
    if_stmt = dynamic.statements[0]
    assert isinstance(if_stmt, NodeIfStmt)
    assert len(if_stmt.clauses) == 3  #if statement has 3 clauses
    clause_a_1 = if_stmt.clauses[0]
    clause_a_2 = if_stmt.clauses[1]
    clause_a_3 = if_stmt.clauses[2]
    #each clause contains one assignment
    assert len(clause_a_1.statements) == 1
    assert isinstance(clause_a_1.statements[0], NodeAssignment)
    assert len(clause_a_2.statements) == 1
    assert isinstance(clause_a_2.statements[0], NodeAssignment)
    assert len(clause_a_3.statements) == 1
    assert isinstance(clause_a_3.statements[0], NodeAssignment)
    #the last clause is an else statement: condition equivalent to True, constant 
    assert isinstance(clause_a_3.condition, (IFloat, IBool))
    assert isrole(clause_a_3.condition, RoleConstant)
    assert bool(clause_a_3.condition.value) is True



def test_function_type_spec_1(): #IGNORE:C01111
    msg = '''
    Test type specifications for function arguments.
    It must be possible to use the class-name for type specifications of the 
    class' methods.
    
    Test for fixed bug #391386 - https://bugs.launchpad.net/freeode/+bug/391386
    '''
    #skip_test(msg)
    print msg

    from freeode.interpreter import Interpreter    
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    func foo(this, x:A):  
        pass
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')
 


def test_isinstance_1(): #IGNORE:C01111
    msg = '''
    Test the builtin istype(...) function, and
    internal conversion of return values bool -> IBool.
    '''
    #skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter

    prog_text = \
'''
data s: String
data f: Float
data b1,b2,b3,b4: Bool

b1 = istype(s, String)
b2 = istype(s, Float)
b3 = istype(f, String)
b4 = istype(f, Float)
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')

    #the module
    mod = intp.modules['test']
    #print mod
    
    #look at variables
    assert mod.b1.value == True
    assert mod.b2.value == False
    assert mod.b3.value == False
    assert mod.b4.value == True



def test_replace_attr_1(): #IGNORE:C01111
    msg = '''
    Test the builtin replace_attr(...) function.
    
    replace_attr(...) is really meant to enable __diff__ for user defined
    classes. 
    #TODO test associate_state_dt(...)
    '''
    skip_test(msg)
    print msg
    
    from freeode.interpreter import Interpreter, IFloat
    #from freeode.util import aa_make_tree  #pylint:disable-msg=W0612 


    prog_text = \
'''
class A:
    data f: Float

#create the variables especially a.f, f_new
data a: A
data f_new: Float

#replace the original a.f and put f_new into its place
replace_attr(a.f, f_new)
'''

    #interpret the program
    intp = Interpreter()
    intp.interpret_module_string(prog_text, None, 'test')

    #the module
    mod = intp.modules['test']
    #print mod
    
    #look at variables
    a = mod.get_attribute('a')
    a_f = a.get_attribute('f')
    f_new = mod.get_attribute('f_new')
    assert f_new is a_f
    assert isinstance(f_new, IFloat)



if __name__ == '__main__':
    # Debugging code may go here.
    test_Interpreter_assign_emit_code_2()
    pass #pylint: disable-msg=W0107

